import argparse
import configparser

import logging
from .logging import *
from .configfiles import stdConfigPath
from .ldap import LDAPConnection,LDAPObjects
from . import __version__
import ldap3

log = logging.getLogger('quadrat2.groups')

class POSIXGroups(LDAPObjects):
    OBJ = ['posixGroup']
    def __call__(self,gid):
        reader = ldap3.Reader(self.ldap.connection,self.obj,
                              'ou=Group,'+self.base,'cn:={}'.format(gid))
        reader.search()
        if len(reader)<1:
            raise RuntimeError('no such group {}'.format(gid))
        return reader[0]

    def create(self,group,gid):
        reader = ldap3.Reader(self.ldap.connection,self.obj,
                              'ou=Group,'+self.base,'cn:={}'.format(group))
        reader.search()
        if len(reader)>0:
            raise RuntimeError ('group {} already exists'.format(group))
        reader = ldap3.Reader(self.ldap.connection,self.obj,
                              'ou=Group,'+self.base,'gidNumber:={}'.format(gid))
        reader.search()
        if len(reader)>0:
            raise RuntimeError ('group with gid {} already exists'.format(gid))

        obj = ldap3.ObjectDef(self.OBJ, self.ldap.connection)
        writer = ldap3.Writer(self.ldap.connection, obj)
        entry = writer.new('cn={},ou=Group,{}'.format(group,self.base))
        entry.gidNumber = gid
        return entry

        

def main():
    CFG = stdConfigPath('quadrat2')
    
    parser = argparse.ArgumentParser(parents=[loggingParser()],
                                     description='Manage groups.')
    
    
    parser.add_argument('group',help="The name of the group")
    parser.add_argument('members', metavar='USER',nargs='*',help="the group members")

    act = parser.add_mutually_exclusive_group()
    act.add_argument("-a", "--add",nargs='+',metavar='USER',                     
                     help="Add USER to the group")

    act.add_argument("-r", "--remove", nargs='+',metavar='USER',   
                   help="Remove USER from the group.")

    act.add_argument("-C", "--create",metavar='GID',type=int,
                   help="Create the group with GID")

    act.add_argument("-D", "--delete",
                   action="store_true", default=False,
                   help="Delete the entire group.")

    control = parser.add_argument_group("Control program behaviour")
    control.add_argument('-k','--keytab', 
                         action="store_true", default=False,
                         help="use host keytab to get krb5 credentials")
    control.add_argument("-y", "--yes",
                         action="store_true", default=False,
                         help="Never prompt for confirmation.")
    control.add_argument("-n", "--dry-run", "--no",
                         action="store_true", default=False,
                         help="Display what would be done, but make no changes.")


    parser.add_argument('--version',action='version',version=__version__)

    parser.add_argument('-c','--config',default=CFG,help='read configuration from file, default={}'.format(CFG))

    args = parser.parse_args()

    set_up_logging(args)

    # configure
    config = configparser.SafeConfigParser()
    config.read(args.config)

    for ldap in ('ldap', 'base'):
        setattr(args, ldap, config.get('ldaplist', ldap))

    ldap = LDAPConnection(dry_run=args.dry_run,yes=args.yes)
    if args.keytab:
        ldap.kinit()
    ldap.connect(args.ldap)

    groups = POSIXGroups(ldap,args.base)

    if args.create is not None:
        # create new group
        try:
            grp = groups.create(args.group,args.create)
        except Exception as e:
            log.error(e)
            return
    else:
        try:
            grp = groups(args.group)
        except RuntimeError as e:
            log.error(e)
            return
        
        if args.add is not None:
            # add group members
            grp = grp.entry_writable()
            for m in args.add:
                if m not in grp.memberUid:
                    grp.memberUid += m
        elif args.remove is not None:
            # remove group members
            grp = grp.entry_writable()
            for m in args.remove:
                if m not in grp.memberUid:
                    log.warning('member {} not in group {}'.format(args.group,m))
                    continue
                grp.memberUid -= m
        elif args.delete:
            # delete group
            grp = grp.entry_writable()
            grp.entry_delete()
        else:
            members = grp.memberUid.values
            members.sort()
            log.info('{0.cn}:{0.gidNumber}'.format(grp))
            for m in members:
                log.info('{} {}'.format(' '*len(grp.cn.value),m))

    if grp.entry_status == "Pending changes":
        if ldap.confirm('{}\nmodify {} [y/N]: '.format(str(grp),args.group)):
            grp.entry_commit_changes()
        if ldap.connection.result['result'] != 0:
            log.error("{0[description]} - {0[message]}".format(ldap.connection.result))
    elif grp.entry_status == "Ready for deletion":
        if ldap.confirm('delete group {} [y/N]: '.format(args.group)):
            grp.entry_commit_changes()
        if ldap.connection.result['result'] != 0:
            log.error("{0[description]} - {0[message]}".format(ldap.connection.result))

if __name__ == '__main__':
    main()
