import argparse
import configparser
import pathlib

import sys, getpass
import xattr
import pwd,grp
import shutil
import logging
from .logging import *
from .configfiles import stdConfigPath
from .ldap import LDAPConnection
from .users import CentralUsers, User
from .convert_size import convert_size
from . import __version__

log = logging.getLogger('quadrat2.quota')

def getCEPHFSdu(path):
    result = {}
    du = shutil.disk_usage(path)
    result['total'] = du.total
    result['used'] = du.used
    result['free'] = du.free
    result['path'] = path
    try:
        result['user'] = pwd.getpwuid(path.stat().st_uid).pw_name
    except:
        pass
    try:
        result['group'] = grp.getgrgid(path.stat().st_gid).gr_name
    except:
        pass
    try:
        result['cephfs_max_bytes'] = int(xattr.getxattr(str(path),"ceph.quota.max_bytes"))
    except:
        pass
    for a in ['rentries','rbytes']:
        try:
            result[f'cephfs_{a}'] = int(xattr.getxattr(str(path),f'ceph.dir.{a}'))
        except:
            pass



    return result

def main():
    CFG = stdConfigPath('quadrat2')
    MAP = stdConfigPath('geosmntpnts')
    USER = getpass.getuser()

    parser = argparse.ArgumentParser(parents=[loggingParser()],
                                     prog='quota',
                                     description='display user storage usage')
    parser.add_argument('-u','--user',metavar='USER',
                        default = USER,
                        help='display all storage usage for USER')
    parser.add_argument('path',metavar='PATH',type=pathlib.Path,nargs='?',help='display storage usage for PATH')
    
    parser.add_argument('--version',action='version',version=__version__)

    parser.add_argument('-m','--map-config',default=MAP,help='quota name to mount point map, defaul={}'.format(MAP))
    parser.add_argument('-c','--config',default=CFG,help='read configuration from file, default={}'.format(CFG))

    args = parser.parse_args()

    set_up_logging(args)

    # configure
    config = configparser.ConfigParser()
    config.read(args.config)
    config.read(args.map_config)

    for ldap in ('ldap', 'base'):
        setattr(args, ldap, config.get('ldaplist', ldap))

    data = []
    if args.path is None:
        ldap = LDAPConnection()
        ldap.connect(args.ldap)
        central = LDAPConnection()
        central.connect(config.get('ldaplist', 'centralldap'),anonymous=True)
        cbase = config.get('ldaplist', 'centralbase')
        cUsers = CentralUsers(central,cbase)

        cu = cUsers(args.user)
        user = User(ldap,args.base,cu)
        for v in user.volumes:
            try:
                m = config.get('map',v)
            except:
                log.warning('cannot find mount point for {}'.format(v))
                m = None
            d = getCEPHFSdu(pathlib.Path(m,args.user))
            d['quota'] = user.get_quota(v)
            data.append(d)
    else:
        if not args.path.exists():
            parser.error(f'no such path {args.path}')
        data.append(getCEPHFSdu(args.path))

    # find max length of path
    mlength = 0
    have_quota = False
    for d in data:
        mlength = max(mlength,len(str(d['path'])))
        if 'quota' in d:
            have_quota = True
    outpath = '{:%d}'%(mlength)

    print (outpath.format('path'),end='')
    if have_quota:
        print (' {:>7}'.format('quota'),end='')
    print (' {:>7} {:>7} {:>7} {:>7} {:>7} {}'.format(
        'total','used','free','percent','size','num_files'))

    for d in data:
        print (outpath.format(str(d['path'])),end='')
        if have_quota:
            if 'quota' in d:
                print (' {:>7}'.format(convert_size(d['quota'],inverse=True)),end='')
            else:
                print (' '*8)        
        print (' {:>7} {:>7} {:>7} {:>6.2f}%'.format(
            convert_size(d['total'],inverse=True),
            convert_size(d['used'],inverse=True),
            convert_size(d['free'],inverse=True),
            d['used']/d['total']*100            
        ),end='')
        if 'cephfs_rbytes' in d:
            print(' {:>7} {}'.format(
                convert_size(d['cephfs_rbytes'],inverse=True),
                d['cephfs_rentries']
            ))
        else:
            print ()


if __name__ == '__main__':
    main()
